#!/usr/bin/perl
#
# exilist - small exim list manager
# 
# (c) Tom Kistner aka duncanthrax 2002-?
#
# Licensed under the GNU GPL
# see LICENSE for details
#
#
use strict;
use Sys::Syslog qw(:DEFAULT setlogsock);
use MIME::Parser;
use MIME::WordDecoder;
use exilist;

# ---------------------------------------
my $version = "0.3";
my $debug = 0;
# ---------------------------------------

dlog("exilist version $version");

# ---------------------------------------------------------------------
# main entry point

my $list_id     = $ARGV[0] || '';
my $list_action = $ARGV[1] || '';
my $list_extra  = $ARGV[2] || '';
my $message_id  = $ARGV[3] || '';
my $env_sender_addr = $ARGV[4] || '[empty envelope sender]';
my $hdr_sender_addr = get_email_addr($ARGV[5]) || '[empty header sender]';

if ($list_id) {
    
    dlog("list_id=$list_id list_action=$list_action list_extra=$list_extra");
    
    my $list = get_list($list_id);
    
    dlog("list is $list->{name}\@$list->{domain} (id $list->{id})");

    # read message from exim
    my ($msgref,$msgsize) = read_message(\*STDIN);
    
    if ($msgref) {
      # bounces
      if ($list_action eq 'return') {
        log_as_bounce($list,$list_extra,$msgsize);
      }
      else {
        # detect loops
        unless (loop_detect($list,$msgref)) {
          # get sender of the message
          my $sender = get_user(get_email_addr($env_sender_addr));
          
          unless (dv($sender)) {
            # we don't know the envelope sender, try the header one
            $sender = get_user(get_email_addr($hdr_sender_addr));
          };

          dlog("sender is $sender->{address} (id $sender->{id})");

          # determine if posting is ok
          my $post_ok = 0;
          
          if (dv($sender)) {
            if ($list->{posting_type} == 1) {     # moderators/owner + members
              if ( ($sender->{lists}->{$list->{id}}->{user_subscription} == 2) || # member
                   ($sender->{lists}->{$list->{id}}->{user_status} > 0) ) {       # owner or moderator
                $post_ok = 1;
              };
            }
            elsif ($list->{posting_type} == 0) {  # moderators/owner only
              if ($sender->{lists}->{$list->{id}}->{user_status} > 0) {
                $post_ok = 1;
              };
            }
            else { $post_ok = 1; };               # open list
          }
          else {
            $post_ok = 1 if ($list->{posting_type} == 2);  # unknown sender, but list open
          };
  
          if ($post_ok) {
            my ($kbytes,$num_sent) = send_to_list($list,$message_id,$msgref,$msgsize);
            if ($kbytes) {
              # all ok, update list stats
              update_stats('lists',$list->{id},{ 'stats_kb' => $kbytes , 'stats_postings' => 1 , 'stats_sent' => $num_sent });
              update_stats('users',$sender->{id}, { 'stats_postings' => 1 });
              slog("$message_id post from $sender->{address} (id $sender->{id}) to $list->{name}\@$list->{domain} (id $list->{id}), $num_sent copies with $kbytes kbytes total.");
            }
            else {
              slog("$message_id error during BSMTP delivery, not all users on the list may have received the message.");
              return_error(1,"$message_id error during BSMTP delivery, not all users on the list may have received the message.");
            };
          }
          else {
            slog("$message_id posting not allowed for '$env_sender_addr' (or '$hdr_sender_addr') on '$list->{name}\@$list->{domain}' list.");
            return_error(1,"$message_id posting not allowed for '$env_sender_addr' (or '$hdr_sender_addr') on '$list->{name}\@$list->{domain}' list.");
          };
        }
        else {
          slog("$message_id detected loop on '$list->{name}\@$list->{domain}' list, discarding.");
          # do not bounce this one, fuckup may happen
          return_error(0,"$message_id detected loop on '$list->{name}\@$list->{domain}' list, discarding.");
        };
        
      };
    }
    else {
      # reading message from exim failed
      slog("$message_id error while reading message from exim.");
      return_error(1,"$message_id error while reading message from exim.");
    };

};

# ---------------------------------------------------------------------
# returns error to exim to include in bounce message
# $_[0] -> exit code $_[1] -> output to return to exim
# does not return
sub return_error {
  
  dlog("return_error()");
  
  print $_[1]."\n" if ($_[0] > 0);
  exit($_[0]);
};


# ---------------------------------------------------------------------
# look for our own list header
# $_[0] -> listref $_[1] -> msgref
# returns 1 if we find it
sub loop_detect {
  
  dlog("loop_detect()");
  
  my $list = shift;
  my $msgref = shift;
  
  foreach my $line (@{ $msgref }) {
    # break on headers end
    last if ($line eq "\n");
    # return 1 when we find our own list header
    if ($line =~ /^\QList-Id: <$list->{name}\E\@\Q$list->{domain}>\E/i) {
      return 1;
    };
  };
  
  return 0;
};


# ---------------------------------------------------------------------
# read_message($fh)
# reads a message from $fh and returns an array ref to it and its size in kbytes
#
sub read_message {
  
  dlog("read_message()");
  
  my $fh = shift;
  my @message = ();

  # read the message
  my $bytes = 0;
  my $headers_over = 0;
  # FIXME: check for sigpipe and return undef if it appears
  while (<$fh>) {
    # skip non-headerish lines (the 'From' comes to mind)
    $headers_over = 1 if ($_ eq "\n");
    unless ($headers_over) {
      next unless ( ($_ =~ /^[^ \:]+\:/) || ($_ =~ /^\t/) || ($_ =~ /^ /) );
    };
    # rewrite single-dot lines to double-dot (FIXME ?)
    $_ = "..\n" if ($_ eq ".\n");
    
    $bytes += length($_);
    push @message, $_;
  };
  
  my $kbytes = int( $bytes / 1024) + 1;
  
  $kbytes++ unless ($kbytes);
  
  return (\@message,$kbytes);
};


# ---------------------------------------------------------------------
# send_to_list($cfg_ref,$message_id,$msgref)
# send out a list message
# returns 0 when error occurs or 1 on success
sub send_to_list {

  dlog("send_to_list()");

  my $list = shift;
  my $message_id = shift;
  my $msgref = shift;
  my $msgsize = shift || 0;

  # add our mailing list header
  $msgref = add_header($msgref, "List-Id: <".$list->{name}."@".$list->{domain}.">");

  # add list tag to the subject, if configured
  $msgref = add_subject_tag($msgref, $list->{subject_tag})
    if ($list->{subject_tag});

  # get all users on list
  my $list_users = get_users($list->{id});

  # pump out all copies
  my $kbytes = 1;
  my $num_sent = 0;
  my $bsmtp_error_file = $config->{home}."/temp/".time()."_bsmtp_stderr_output_".$message_id;
  dlog("opening exim as $config->{exim} -bS");
  open(EXIM,"| $config->{exim} -bS 2> $bsmtp_error_file > /dev/null") or dlog("$message_id Error - cannot open exim for BSMTP input");
  print EXIM "HELO exilist\n";
  foreach my $list_user (@{ $list_users }) {
    next if ($list_user->{user_subscription} < 2);  # skip non-subscribed users
    print EXIM "RSET\n";
    print EXIM 'MAIL FROM:<'.$list->{id}.'_return_'.$list_user->{userdata}->{id}.'.'.$message_id.'@'.$list->{domain}.'>'."\n";
    print EXIM 'RCPT TO:<'.$list_user->{userdata}->{address}.'>'."\n";
    print EXIM 'DATA'."\n";
    print EXIM @{ $msgref };
    print EXIM "\n.\n";
    # add stats for user
    update_stats('users',$list_user->{userdata}->{id},{ 'stats_kb' => $msgsize , 'stats_sent' => 1 });
    $kbytes += $msgsize;
    $num_sent++;
  };
  print EXIM "QUIT\n";
  close(EXIM);
  
  # check if any errors occured
  if (-s $bsmtp_error_file) {
    # some error occured during BSMTP delivery
    slog("$message_id error during BSMTP delivery to exim. See '$bsmtp_error_file' for details.");
    return (0,0);
  }
  else {
    unlink $bsmtp_error_file;
    return ($kbytes,$num_sent);
  };
  
};

# ---------------------------------------------------------------------
# add_subject_tag($msgref,$tag)
# add subject to the messages subject
#
sub add_subject_tag {

  dlog("add_subject_tag()");

  my $msgref = shift;
  my $tag = shift;
  
  # find and mangle subject
  # takes care of possible MIME encodings
  LINELOOP: foreach (@{ $msgref }) {
    # skip if already there ..
    eval { next LINELOOP if ($_ =~ /^Subject\: .*(\Q$tag\E).*$/); };
    dlog($@) if $@;
    s/^(Subject\: )(.+)$/$1.$tag." ".unmime($2)/e;
    last if ($_ eq "\n");
  };

  return $msgref;
};


# ---------------------------------------------------------------------
# add_header($msgref,$header)
# add a header to a message
#
sub add_header {
  
  dlog("add_header()");
  
  my $msgref = shift;
  my $header = shift;
  
  my $i = 0;
  foreach (@{ $msgref }) {
    my $hline = $_;
    $hline =~ s/([\r\n]+)//;
    my $newline = $1;
    unless ($hline) {
      splice @{ $msgref }, $i, 1, ($header.$newline,$newline);
     last;
    };
    $i++;
  };
  
  return $msgref;
};


# ---------------------------------------------------------------------
# log_as_bounce($listref,$list_extra,$msgref)
# records a bounce
#
sub log_as_bounce {
  
  my $list = shift;
  my $list_extra = shift;
  my $msgsize = shift;
  
  my $rcpt = '';
  my $old_message_id = '';
  
  if ($list_extra =~ m/([0-9]+)\.([^.]+)/) {
    $rcpt = $1;
    $old_message_id = $2;
  };
  
  unless ($rcpt && $old_message_id) {
    slog("received bounce from malformed return path fragment ($list_extra).");  
  }
  else {
    slog("$old_message_id bounce from user id $rcpt on '$list->{name}'.");
    # try to get user and log bounce
    my $user = get_user($rcpt,0);
    
    if (dv($user)) {
      # record bounce on list and user
      update_stats('lists', $list->{id}, { 'stats_bounced' => 1, 'stats_kb' => $msgsize } );
      update_stats('users', $user->{id}, { 'stats_bounced' => 1, 'stats_kb' => $msgsize } );
    };
  };
};


# ---------------------------------------------------------------------
# slog($message)
# send $message to syslog (level info, facility mail)
#
sub slog {
  setlogsock 'unix';
  openlog 'exilist', 'pid', 'mail';
  syslog 'info', $_[0];
  closelog;
};


# ---------------------------------------------------------------------
# dlog($message)
# send $message to syslog (level info, facility mail)
# if debugging is on
sub dlog {
  if ($debug) {
    setlogsock 'unix';
    openlog 'exilist', 'pid', 'mail';
    syslog 'info', $_[0];
    closelog;
  };
};

